package ui;

###############################################################################
#                                                                             #
# Slrnconf - a graphical configuration utility for SLRN                       #
# Copyright (C) 2000-2003 Torsten Schoenfeld                                  #
#                                                                             #
# This program is free software; you can redistribute it and/or modify it     #
# under the terms of the GNU General Public License as published by the Free  #
# Software Foundation; either version 2 of the License, or (at your option)   #
# any later version.                                                          #
#                                                                             #
# This program is distributed in the hope that it will be useful, but WITHOUT #
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       #
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   #
# more details.                                                               #
#                                                                             #
# You should have received a copy of the GNU General Public License along     #
# with this program; if not, write to the Free Software Foundation, Inc., 59  #
# Temple Place, Suite 330, Boston, MA 02111-1307 USA                          #
#                                                                             #
###############################################################################

use strict;
use Gtk2 -init;

use preferences;

sub new {
  my ($class, $root, $help, $data) = @_;
  my $self = {};

  bless($self, $class);

  $self -> { root } = $root;
  $self -> { help } = $help;
  $self -> { data } = $data;

  $self -> { preferences } = preferences -> new($self);

  return $self;
}

sub bitmap_popup {
  my ($self, $caption, @variable) = @_;
  my $value = $self -> { preferences } -> get(@variable);

  my %bitmap = $self -> get_bitmap($variable[-1]);
  my %config = ();

  #############################################################################

  foreach (sort { $b <=> $a } (keys(%bitmap))) {
    if ($value >= $_) {
      $value -= $_;
      $config{ $_ } = 1;
    }
    else {
      $config{ $_ } = 0;
    }
  }

  #############################################################################

  my $dialog = Gtk2::Dialog -> new($caption,
                                   $self -> { main_window },
                                   [qw(modal destroy_with_parent)],
                                   "gtk-ok" => "ok",
                                   "gtk-cancel" => "cancel");

  #############################################################################

  foreach my $option(sort { $a <=> $b } (keys(%bitmap))) {
    my $button = Gtk2::ToggleButton -> new(" $bitmap{ $option } ");

    $button -> set_active($config{ $option });

    $button -> signal_connect(toggled => sub {
      $config{ $option } = $button -> get_active();
    });

    $dialog -> vbox -> pack_start($button, 1, 1, 0);
  }

  #############################################################################

  $dialog -> show_all();

  #############################################################################

  my $response = $dialog -> run();
  $dialog -> destroy();

  if ($response eq "ok") {
    my $return_value = 0;

    foreach (keys(%config)) {
      $return_value += $_ if ($config{ $_ });
    }

    return $return_value;
  }

  return;
}

sub create_mainwindow {
  my ($self) = @_;

  $self -> { main_window } = Gtk2::Window -> new("toplevel");
  $self -> { main_box } = Gtk2::VBox -> new(0, 0);

  $self -> { accel_group } = Gtk2::AccelGroup -> new();

  $self -> { main_window } -> set_title("Slrnconf");

  $self -> { main_window } -> signal_connect(delete_event => sub {
    $self -> quit();
    return 1;
  });

  return 1;
}

sub create_menu {
  my ($self) = @_;

  my $item_factory = Gtk2::ItemFactory -> new("Gtk2::MenuBar",
                                              "<Slrnconf>/Main",
                                              $self -> { accel_group });

  $item_factory -> create_items(undef,
    {
      path => "/_File",
      item_type => "<Branch>"
    },
    {
      path => "/File/_Open",
      callback => sub { $self -> file_open(); },
      item_type => "<StockItem>",
      extra_data => "gtk-open"
    },
    {
      path => "/File/_Save",
      callback => sub { $self -> file_save(); },
      item_type => "<StockItem>",
      extra_data => "gtk-save"
    },
    {
      path => "/File/Save _As",
      callback => sub { $self -> file_save_as(); },
      item_type => "<StockItem>",
      extra_data => "gtk-save-as"
    },
    {
      path => "/File/Sep1",
      item_type => "<Separator>"
    },
    {
      path => "/File/_Quit",
      callback => sub { $self -> quit(); },
      item_type => "<StockItem>",
      extra_data => "gtk-quit"
    },
    {
      path => "/_Help",
      item_type => "<LastBranch>"
    },
    {
      path => "/Help/_About",
      callback => sub { $self -> show_file($self -> { root } . "/ABOUT"); }
    },
    {
      path => "/Help/_License",
      callback => sub { $self -> show_file($self -> { root } . "/COPYING"); }
    },
    {
      path => "/Help/Sep2",
      item_type => "<Separator>"
    },
    {
      path => "/Help/_Help",
      callback => sub { $self -> show_file($self -> { root } . "/README"); },
      item_type => "<StockItem>",
      extra_data => "gtk-help"
    }
  );

  $self -> { main_box } -> pack_start(
    $item_factory -> get_widget("<Slrnconf>/Main"), 0, 0, 0);

  return 1;
}

sub create_notebook {
  my ($self) = @_;

  $self -> { notebook } = Gtk2::Notebook -> new();
  $self -> { notebook } -> set_tab_pos("top");

  $self -> { main_box } -> pack_start($self -> { notebook }, 1, 1, 0);

  return 1;
}

sub create_tabs {
  my ($self) = @_;

  my @tabs = (
    ["general",        "General"],
    ["group",          "Groups"],
    ["article",        "Articles"],
    ["post_and_mail",  "Post & Mail"],
    ["misc",           "Misc"],
    ["files_and_dirs", "Files & Dirs"],
    ["scoring",        "Scoring"],
    ["spool",          "Spool"],
    ["grouplens",      "Grouplens"],
    ["color",          "Color"],
    ["mono",           "Mono"]
  );

  foreach (@tabs) {
    my ($tab_id, $tab_label) = @$_;

    my @fields = $self -> get_tab($tab_id);

    my $container = Gtk2::ScrolledWindow -> new();
    my $table = Gtk2::Table -> new($#fields, 3, 0);

    my $current_row = 0;

    if ($tab_id eq "color") {
      $table -> attach(Gtk2::Label -> new("Foreground"), 1, 2, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);
      $table -> attach(Gtk2::Label -> new("Background"), 2, 3, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);

      my $button = Gtk2::Button -> new(" ? ");

      $button -> signal_connect(clicked => sub {
        $self -> show_file($self -> { help } . "/color");
      });

      $table -> attach($button, 3, 4, $current_row, $current_row + 1, "shrink",
                       "shrink", 5, 0);

      $current_row++;
    }

    elsif ($tab_id eq "mono") {
      $table -> attach(Gtk2::Label -> new("1st"), 1, 2, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);
      $table -> attach(Gtk2::Label -> new("2nd"), 2, 3, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);
      $table -> attach(Gtk2::Label -> new("3rd"), 3, 4, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);
      $table -> attach(Gtk2::Label -> new("4th"), 4, 5, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);

      my $button = Gtk2::Button -> new(" ? ");

      $button -> signal_connect(clicked => sub {
        $self -> show_file($self -> { help } . "/mono");
      });

      $table -> attach($button, 5, 6, $current_row, $current_row + 1, "shrink",
                       "shrink", 5, 0);

      $current_row++;
    }

    foreach (@fields) {
      my $caption = shift(@$_);

      #########################################################################

      if ($caption eq "<seperator>") {
        my $separator = Gtk2::HSeparator -> new();

        $separator -> set("height-request" => 20);

        $table -> attach($separator, 0, 3, $current_row, $current_row + 1,
                         ["fill", "expand"], "shrink", 0, 0);

        $current_row++;

        next;
      }

      #########################################################################

      my @variable = split(/, /, shift(@$_));
      my $type = shift(@$_);

      my $label = Gtk2::Label -> new($caption);

      $label -> set_alignment(1, 0);

      $table -> attach($label, 0, 1, $current_row, $current_row + 1, "fill",
                       "shrink", 5, 0);

      #########################################################################

      if ($type =~ m/^\*?text$/) {
        my $entry = Gtk2::Entry -> new();

        my $signal_id = $entry -> signal_connect(changed => sub {
          $self -> { preferences } -> set(@variable, shift() -> get_text());
        });

        $self -> { preferences } ->
          attach_callback(@variable, sub {
            my ($value) = @_;

            $value = "" unless (defined($value));

            $entry -> signal_handler_block($signal_id);
            $entry -> set_text($value);
            $entry -> signal_handler_unblock($signal_id);
          });

        if ($type eq "*text") {
          my $hbox = Gtk2::HBox -> new(0, 0);
          my $button = Gtk2::Button -> new(" ... ");

          $button -> signal_connect(clicked => sub {
            $self -> file_dialog("open", $entry);
          });

          $hbox -> pack_start($entry, 1, 1, 0);
          $hbox -> pack_start($button, 0, 0, 0);

          $table -> attach($hbox, 1, 2, $current_row, $current_row + 1,
                           ["fill", "expand"], "shrink", 0, 0);
        }
        else {
          $table -> attach($entry, 1, 2, $current_row, $current_row + 1,
                           ["fill", "expand"], "shrink", 0, 0);
        }
      }

      #########################################################################

      elsif ($type eq "check") {
        my $check = Gtk2::CheckButton -> new();

        my $signal_id = $check -> signal_connect(toggled => sub {
          $self -> { preferences } -> set(@variable,
                                          shift() -> get_active() ? 1 : 0);
        });

        $self -> { preferences } ->
          attach_callback(@variable, sub {
            my ($value) = @_;

            unless (defined($value)) {
              $value = 0;
            }
            elsif ($value != 0 && $value != 1) {
              $value = 1;
            }

            $check -> signal_handler_block($signal_id);
            $check -> set_active($value);
            $check -> signal_handler_unblock($signal_id);
          });

        $table -> attach($check, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "shrink", 0, 0);
      }

      #########################################################################

      elsif ($type eq "menu") {
        my $item_factory = Gtk2::ItemFactory -> new("Gtk2::OptionMenu",
                                                    "<Slrnconf>/Menu/$variable[-1]");

        my @menu = $self -> get_menu($variable[-1]);

        $item_factory -> create_items(undef,
          map {
            my ($label, $key) = @$_;

            $label =~ s/_/__/g;

            {
              path => "/" . $label,
              callback => sub {
                $self -> { preferences } -> set(@variable, $key);
              }
            }
          } @menu
        );

        my $optionmenu =
          $item_factory -> get_widget("<Slrnconf>/Menu/$variable[-1]");

        $optionmenu -> set("height-request" => 25);

        $self -> { preferences } -> attach_callback(@variable, sub {
                                      my ($value) = @_;

                                      if (defined($value)) {
                                        foreach (0 .. $#menu) {
                                          if ($menu[$_] -> [1] eq $value) {
                                            $optionmenu -> set_history($_);
                                          }
                                        }
                                      }
                                    });

        $table -> attach($optionmenu, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "fill", 0, 0);
      }

      #########################################################################

      elsif ($type eq "bitmap") {
        my $value = $self -> { preferences } -> get(@variable);
        $value = 0 unless (defined($value));

        my $button = Gtk2::Button -> new($value);

        $button -> signal_connect(clicked => sub {
          my $new_value = $self -> bitmap_popup($caption, @variable);

          if (defined($new_value)) {
            $self -> { preferences } -> set(@variable, $new_value);
            shift() -> set_label($new_value);
          }
        });

        $self -> { preferences } -> attach_callback(@variable, sub {
                                      my ($value) = @_;

                                      if (defined($value)) {
                                        $button -> set_label($value);
                                      }
                                    });

        $table -> attach($button, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "fill", 0, 0);
      }

      #########################################################################

      elsif ($type eq "list") {
        my $dialog_title = shift(@$_);
        my $dialog = shift(@$_);

        my $container = Gtk2::HBox -> new(0, 0);

        #######################################################################

        my $list_container = Gtk2::ScrolledWindow -> new();
        my $model = Gtk2::ListStore -> new("Glib::String");
        my $view = Gtk2::TreeView -> new($model);

        my $cell_renderer = Gtk2::CellRendererText -> new();
        my $view_column =
          Gtk2::TreeViewColumn -> new_with_attributes("Value",
                                                      $cell_renderer,
                                                      text => 0);

        $view -> append_column($view_column);

        $self -> { preferences } -> attach_callback(@variable, sub {
                                      my ($value) = @_;

                                      if (defined($value)) {
                                        $model -> clear();

                                        foreach (@{$value}) {
                                          $model -> set($model -> append(),
                                                        0 => $_);
                                        }
                                      }
                                    });

        $view -> set_headers_visible(0);
        $view -> get_selection() -> set_mode("single");

        $view -> signal_connect(row_activated => sub {
          my ($view, $path, $view_column) = @_;

          Glib::Idle -> add(sub {
            $self -> manage_list($view, "edit", \@variable, $dialog_title,
                                 $dialog);
            return 0;
          });
        });

        $view -> signal_connect(button_release_event => sub {
          my ($view, $event) = @_;

          if ($event -> button == 3) {
            $self -> popup($view, \@variable, $dialog_title, $dialog);
            return 1;
          }

          return 0;
        });

        $list_container -> add_with_viewport($view);
	$list_container -> set("height-request" => 60);
        $list_container -> set_policy("automatic", "automatic");

        #######################################################################

        my $button_box = Gtk2::VBox -> new(1, 0);

        my $arrow_up = Gtk2::Arrow -> new("up", "none");
        my $arrow_down = Gtk2::Arrow -> new("down", "none");

        my $button_up = Gtk2::Button -> new();
        my $button_down = Gtk2::Button -> new();

        $button_up -> add($arrow_up);
        $button_down -> add($arrow_down);

        $button_up -> signal_connect(clicked => sub {
          $self -> manage_list($view, "up", \@variable);
        }),

        $button_down -> signal_connect(clicked => sub {
          $self -> manage_list($view, "down", \@variable);
        }),

        $button_box -> pack_start($button_up, 1, 1, 0);
        $button_box -> pack_start($button_down, 1, 1, 0);

        #######################################################################

        $container -> pack_start($list_container, 1, 1, 0);
        $container -> pack_start($button_box, 0, 0, 0);

        $table -> attach($container, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "shrink", 0, 0);
      }

      #########################################################################

      elsif ($type eq "color") {
        my $identifier = $variable[-1];

        foreach my $type (qw(fg bg)) {
          my @menu = $self -> get_colors();

          my $item_factory = Gtk2::ItemFactory -> new("Gtk2::OptionMenu",
                                                      "<Slrnconf>/Color/$identifier/$type");

          $item_factory -> create_items(undef,
            map {
              my $color = $_;

              {
                path => "/" . $color,
                callback => sub {
                  $self -> { preferences } -> set("color",
                                                  $identifier . "_" . $type,
                                                  $color);
                }
              }
            } @menu
          );

          my $optionmenu = $item_factory -> get_widget("<Slrnconf>/Color/$identifier/$type");
          $optionmenu -> set("height-request" => 25);

          $self -> { preferences } ->
            attach_callback("color", $identifier . "_" . $type, sub {
              my ($value) = @_;

              if (defined($value)) {
                foreach (0 .. $#menu) {
                  if ($menu[$_] eq $value) {
                    $optionmenu -> set_history($_);
                  }
                }
              }
            });

          if ($type eq "fg") {
            $table -> attach_defaults($optionmenu, 1, 2, $current_row,
                                      $current_row + 1);
          }
          elsif ($type eq "bg") {
            $table -> attach_defaults($optionmenu, 2, 3, $current_row,
                                      $current_row + 1);
          }
        }
      }

      #########################################################################

      elsif ($type eq "mono") {
        my $identifier = $variable[-1];

        foreach my $type (qw(one two three four)) {
          my @menu = $self -> get_monos();

          my $item_factory = Gtk2::ItemFactory -> new("Gtk2::OptionMenu",
                                                      "<Slrnconf>/Mono/$identifier/$type");

          $item_factory -> create_items(undef,
            map {
              my $mono = $_;

              {
                path => "/" . $mono,
                callback => sub {
                  $self -> { preferences } -> set("mono",
                                                  $identifier . "_" . $type,
                                                  $mono);
                }
              }
            } @menu
          );

          my $optionmenu = $item_factory -> get_widget("<Slrnconf>/Mono/$identifier/$type");
          $optionmenu -> set("height-request" => 25);

          $self -> { preferences } ->
            attach_callback("mono", $identifier . "_" . $type, sub {
              my ($value) = @_;

              if (defined($value)) {
                foreach (0 .. $#menu) {
                  if ($menu[$_] eq $value) {
                    $optionmenu -> set_history($_);
                  }
                }
              }
            });

          if ($type eq "one") {
            $table -> attach_defaults($optionmenu, 1, 2, $current_row,
                                      $current_row + 1);
          }
          elsif ($type eq "two") {
            $table -> attach_defaults($optionmenu, 2, 3, $current_row,
                                      $current_row + 1);
          }
          elsif ($type eq "three") {
            $table -> attach_defaults($optionmenu, 3, 4, $current_row,
                                      $current_row + 1);
          }
          elsif ($type eq "four") {
            $table -> attach_defaults($optionmenu, 4, 5, $current_row,
                                      $current_row + 1);
          }
        }
      }

      #########################################################################

      unless ($tab_id eq "color" || $tab_id eq "mono") {
        my $button = Gtk2::Button -> new(" ? ");

        $button -> signal_connect(clicked => sub {
          $self -> show_file($self -> { help } . "/" . $variable[-1]);
        });

        $table -> attach($button, 2, 3, $current_row, $current_row + 1,
                         "shrink", "shrink", 5, 0);
      }

      $current_row++;
    }

    ###########################################################################

    $table -> set_border_width(5);

    $container -> set_size_request(600, 250);
    $container -> set_policy("automatic", "automatic");
    $container -> set_border_width(1);

    $container -> add_with_viewport($table);

    $self -> { notebook } -> append_page($container,
                                         Gtk2::Label -> new($tab_label));
  }

  return 1;
}

sub create_statusbar {
  my ($self) = @_;

  $self -> { statusbar } = Gtk2::Statusbar -> new();
  $self -> { statusbar_id } = $self -> { statusbar } ->
                                get_context_id("statusbar");

  $self -> { main_box } -> pack_start($self -> { statusbar }, 0, 0, 0);

  $self -> rc_file();

  return 1;
}

sub dialog {
  my ($self, $dialog_title, $dialog, $old_input) = @_;

  my @old_input = split(/ => /, $old_input || "");
  my $input = [];

  #############################################################################

  my @dialog_rows;

  foreach my $label(split(/, /, $dialog)) {
    my $dialog_row = { };

    if ($label =~ s/^\*//) {
      $dialog_row -> { browse } = 1;
    }

    $dialog_row -> { label } = $label;

    push(@dialog_rows, $dialog_row);
  }

  #############################################################################

  my $dialog_box = Gtk2::Dialog -> new($dialog_title,
                                       $self -> { main_window },
                                       [qw(modal destroy_with_parent)],
                                       "gtk-ok" => "ok",
                                       "gtk-cancel" => "cancel");

  #############################################################################

  my $table = Gtk2::Table -> new($#dialog_rows, 3, 0);
  my $row = 0;

  foreach (@dialog_rows) {
    my $entry = Gtk2::Entry -> new();
    my $label = Gtk2::Label -> new($_ -> { label });

    my $local_row = $row;

    $entry -> signal_connect(changed => sub {
      $input -> [$local_row] = shift() -> get_text();
    });

    if (defined($old_input[$row])) {
      $entry -> set_text($old_input[$row]);
    }

    $label -> set_alignment(1, 0.4);

    $table -> attach_defaults($label, 0, 1, $row, $row + 1);
    $table -> attach_defaults($entry, 1, 2, $row, $row + 1);

    if ($_ -> { browse }) {
      my $button_browse = Gtk2::Button -> new(" ... ");

      $button_browse -> signal_connect(clicked => sub {
        $self -> file_dialog("open", $entry);
      });

      $table -> attach_defaults($button_browse, 2, 3, $row, $row + 1);
    }

    $row++;
  }

  #############################################################################

  $table -> set_row_spacings(5);
  $table -> set_col_spacings(5);

  $dialog_box -> vbox -> pack_start($table, 1, 1, 0);

  $dialog_box -> set_border_width(7);
  $dialog_box -> show_all();

  #############################################################################

  my $response = $dialog_box -> run();
  $dialog_box -> destroy();

  return $response eq "ok" ? $input : undef;
}

sub error {
  my ($self, $error) = @_;

  my $id = $self -> { statusbar } -> push($self -> { statusbar_id }, $error);

  Glib::Timeout -> add(10000, sub {
    $self -> { statusbar } -> remove($self -> { statusbar_id }, $id);
    return 0;
  });

  return 0;
}

sub file_dialog {
  my ($self, $action, $entry) = @_;

  #############################################################################

  my $file_selector = Gtk2::FileSelection -> new(
    ($action eq "save") ?
      "Select a file for saving" :
      "Select the file to open");

  $file_selector -> signal_connect(delete_event => sub {
    Gtk2 -> main_quit();
    return 1;
  });

  $file_selector -> ok_button -> signal_connect(clicked => sub {
    my $file = $file_selector -> get_filename();

    if ($action eq "save" && -e $file) {
      return 1
        unless ($self -> question("Do you really want to overwrite '$file'?"));
    }

    if (defined($entry) && ref($entry) eq "Gtk2::Entry") {
      $entry -> set_text($file);
    }
    else {
      $self -> { _tmp_file } = $file;
    }

    Gtk2 -> main_quit();
  });

  $file_selector -> cancel_button -> signal_connect(clicked => sub {
    Gtk2 -> main_quit();
  });

  $file_selector -> set_modal(1);
  $file_selector -> set_transient_for($self -> { main_window });
  $file_selector -> show_all();

  #############################################################################

  Gtk2 -> main();

  $file_selector -> destroy();

  return delete($self -> { _tmp_file });
}

sub file_open {
  my ($self) = @_;

  my $file = $self -> file_dialog("open");

  if (defined($file) && -f $file) {
    $self -> rc_file($file);
    Glib::Idle -> add(sub {
      $self -> read_preferences();
      return 0;
    });
  }

  return 1;
}

sub file_save {
  my ($self) = @_;
  return $self -> { preferences } -> write();
}

sub file_save_as {
  my ($self) = @_;
  my $file = $self -> file_dialog("save");

  if (defined($file)) {

    $self -> rc_file($file);
    $self -> { preferences } -> write();
  }

  return 1;
}

sub finalize_mainwindow {
  my ($self) = @_;

  $self -> { main_window } -> add($self -> { main_box });
  $self -> { main_window } -> add_accel_group($self -> { accel_group });

  $self -> { main_window } -> show_all();

  return 1;
}

sub get_bitmap {
  my ($self, $bitmap) = @_;
  my $indicator = 0;

  my %bitmap = ();

  open(BITMAPS, $self -> { data } . "/bitmaps") or
    die($self -> { data } . "/bitmaps: $!\n");

  while (my $line = <BITMAPS>) {
    if ($line =~ m/^$bitmap$/) {
      $indicator = 1;

      next;
    }

    if ($indicator && $line =~ s/^  //) {
      chomp($line);

      my ($key, $value) = split(/\t+/, $line);

      $bitmap{ $key } = $value;
    }
    else {
      $indicator = 0;
    }
  }

  close(BITMAPS);

  return %bitmap;
}

sub get_colors {
  my ($self) = @_;

  local $/ = undef;

  open(COLORS, $self -> { data } . "/colors") or
    die($self -> { data } . "/colors: $!\n");

  my @colors = split(/\n/, <COLORS>);

  close(COLORS);

  return @colors;
}

sub get_menu {
  my ($self, $menu) = @_;
  my $indicator = 0;

  my ($line, @menu);

  open(MENUS, $self -> { data } . "/menus") or
    die($self -> { data } . "/menus: $!\n");

  while ($line = <MENUS>) {
    if ($line =~ m/^$menu$/) {
      $indicator = 1;
      next;
    }

    if ($indicator && $line =~ s/^  //) {
      chomp($line);
      push(@menu, [split(/\t+/, $line)]);
    }
    else {
      $indicator = 0;
    }
  }

  close(MENUS);

  return @menu;
}

sub get_monos {
  my ($self) = @_;

  local $/ = undef;

  open(MONOS, $self -> { data } . "/monos") or
    die($self -> { data } . "/monos: $!\n");

  my @monos = split(/\n/, <MONOS>);

  close(MONOS);

  return @monos;
}

sub get_tab {
  my ($self, $tab) = @_;
  my $indicator = 0;

  my ($line, @tab);

  open(TABS, $self -> { data } . "/tabs") or
    die($self -> { data } . "/tabs: $!\n");

  while ($line = <TABS>) {
    if ($line =~ m/^$tab$/) {
      $indicator = 1;
      next;
    }

    if ($indicator && $line =~ s/^  //) {
      chomp($line);
      push(@tab, [split(/\t+/, $line)]);
    }
    else {
      $indicator = 0;
    }
  }

  close(TABS);

  return @tab;
}

sub manage_list {
  my ($self, $view, $action, $variable, $dialog_title, $dialog) = @_;
  my $model = $view -> get_model();

  #############################################################################

  if ($action eq "add") {
    my $input = $self -> dialog($dialog_title, $dialog);

    if (defined($input)) {
      $model -> set($model -> append(),
                    0 => join(" => ", @{$input}));
    }
  }

  #############################################################################

  else {
    my $iterator = $view -> get_selection() -> get_selected();

    if (defined($iterator)) {
      if ($action eq "del") {
        $model -> remove($iterator);
      }

      elsif ($action eq "up") {
        my $path = $model -> get_path($iterator);
        if ($path -> prev()) {
          $model -> move_after($model -> get_iter($path), $iterator);
        }
      }

      elsif ($action eq "down") {
        my $path = $model -> get_path($iterator);
        $path -> next();
        $model -> move_after($iterator, $model -> get_iter($path));
      }

      elsif ($action eq "edit") {
        my $old_value = $model -> get($iterator, 0) || return 0;
        my $input = $self -> dialog($dialog_title, $dialog, $old_value);

        if (defined($input)) {
          $model -> set($iterator,
                        0 => join(" => ", @{$input}));
        }
      }
    }
  }

  #############################################################################

  # FIXME: Rather update the changed value directly.  Doing everything
  # everytime sets the internal "changed" flag even if nothing was edited.

  my $i = 0;

  $model -> foreach(sub {
    my ($model, $path, $iterator) = @_;
    $self -> { preferences } -> index_set(@{$variable}, $i++,
                                          $model -> get($iterator, 0));
  });

  #############################################################################

  return 1;
}

sub popup {
  my ($self, $view, $variable, $dialog_title, $dialog) = @_;

  my $item_factory = Gtk2::ItemFactory -> new("Gtk2::Menu",
                                              "<Slrnconf>/Popup");

  if (defined($view -> get_selection() -> get_selected())) {
    $item_factory -> create_items(undef,
      {
        path => "/_Add",
        callback => sub { $self -> manage_list($view, "add", $variable, $dialog_title, $dialog); },
        item_type => "<StockItem>",
        extra_data => "gtk-add"
      },
      {
        path => "/_Edit",
        callback => sub { $self -> manage_list($view, "edit", $variable, $dialog_title, $dialog); },
        item_type => "<StockItem>",
        extra_data => "gtk-find-and-replace",
        accelerator => ""
      },
      {
        path => "/_Delete",
        callback => sub { $self -> manage_list($view, "del", $variable); },
        item_type => "<StockItem>",
        extra_data => "gtk-delete"
      }
    );
  }
  else {
    $item_factory -> create_items(undef,
      {
        path => "/_Add",
        callback => sub { $self -> manage_list($view, "add", $variable, $dialog_title, $dialog); },
        item_type => "<StockItem>",
        extra_data => "gtk-add"
      }
    );
  }

  $item_factory -> get_widget("<Slrnconf>/Popup") -> popup(undef,
                                                           undef,
                                                           undef,
                                                           undef,
                                                           3,
                                                           0);

  return 1;
}

sub question {
  my ($self, $question) = @_;

  my $dialog = Gtk2::MessageDialog -> new($self -> { main_window },
                                          [qw(modal destroy_with_parent)],
                                          "question",
                                          "yes_no",
                                          $question);

  my $response = $dialog -> run();
  $dialog -> destroy();

  return $response eq "yes" ? 1 : 0;
}

sub quit {
  my ($self) = @_;

  if ($self -> { preferences } -> { _changed }) {
    return 1
      unless ($self -> question("Really quit and abandon all changes?"));
  }

  Gtk2 -> main_quit();

  return 1;
}

sub read_preferences {
  my ($self) = @_;

  my $window = Gtk2::Window -> new("popup");
  my $label = Gtk2::Label -> new("Reading " . $self -> rc_file() . " ...");

  $window -> set_transient_for($self -> { main_window });
  $window -> set_default_size(250, 50);
  $window -> set_position("center-on-parent");

  $window -> add($label);
  $window -> show_all();

  $self -> { main_window } -> set_sensitive(0);

  Glib::Timeout -> add(100, sub {
    $self -> { preferences } -> read();
    $window -> destroy();
    $self -> { main_window } -> set_sensitive(1);
    return 0;
  });

  return 1;
}

sub rc_file {
  my ($self, $new_rc_file) = @_;

  if (defined($new_rc_file)) {
    $self -> { preferences } -> { rc_file } = $new_rc_file;
  }

  $self -> { main_window } ->
    set_title("Slrnconf - " . $self -> { preferences } -> { rc_file });

  return $self -> { preferences } -> { rc_file };
}

sub run {
  my ($self) = @_;

  $self -> create_mainwindow();
  $self -> create_menu();
  $self -> create_notebook();
  $self -> create_statusbar();
  $self -> create_tabs();

  $self -> finalize_mainwindow();

  Glib::Idle -> add(sub {
    $self -> read_preferences();
    return 0;
  });

  Gtk2 -> main();

  return 1;
}

sub show_file {
  my ($self, $file) = @_;

  my $window = Gtk2::Window -> new("toplevel");
  my $container = Gtk2::ScrolledWindow -> new();
  my $text_view = Gtk2::TextView -> new();
  my $text_buffer = $text_view -> get_buffer();

  $window -> set_title("Slrnconf - $file");
  $window -> signal_connect(delete_event => sub {
    Gtk2 -> main_quit();
    return 1;
  });

  open(HELP, "<$file") or
    return $self -> error("Couldn't open '$file' for reading.");

  local $/ = undef;
  $text_buffer -> set_text(<HELP>);
  # $text_buffer -> insert_with_tags($text_buffer -> get_start_iter(),
  #                                  <HELP>,
  #                                  $text_buffer -> create_tag("fixed", font => "Fixed"));

  close(HELP);

  $text_view -> set_editable(0);
  $text_view -> modify_font(Gtk2::Pango::FontDescription -> from_string("Fixed"));

  $container -> add($text_view);
  $container -> set_policy("automatic", "automatic");

  $window -> add($container);
  $window -> set_default_size(500, 200);
  $window -> set_modal(1);
  $window -> show_all();

  Gtk2 -> main();

  $window -> destroy();

  return 1;
}

1;
